"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

MQTT = {
    "dsmr_reader": [
        "dsmr/#",
    ],
    "fully_kiosk": [
        "fully/deviceInfo/+",
    ],
    "tasmota": [
        "tasmota/discovery/#",
    ],
}
