"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

SSDP = {
    "arcam_fmj": [
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1",
            "manufacturer": "ARCAM",
        },
    ],
    "axis": [
        {
            "manufacturer": "AXIS",
        },
    ],
    "braviatv": [
        {
            "manufacturer": "Sony Corporation",
            "st": "urn:schemas-sony-com:service:ScalarWebAPI:1",
        },
    ],
    "control4": [
        {
            "st": "c4:director",
        },
    ],
    "deconz": [
        {
            "manufacturer": "Royal Philips Electronics",
            "manufacturerURL": "http://www.dresden-elektronik.de",
        },
    ],
    "denonavr": [
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1",
            "manufacturer": "Denon",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1",
            "manufacturer": "DENON",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1",
            "manufacturer": "DENON PROFESSIONAL",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1",
            "manufacturer": "Marantz",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "manufacturer": "Denon",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "manufacturer": "DENON",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "manufacturer": "DENON PROFESSIONAL",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "manufacturer": "Marantz",
        },
        {
            "deviceType": "urn:schemas-denon-com:device:AiosDevice:1",
            "manufacturer": "Denon",
        },
        {
            "deviceType": "urn:schemas-denon-com:device:AiosDevice:1",
            "manufacturer": "DENON",
        },
        {
            "deviceType": "urn:schemas-denon-com:device:AiosDevice:1",
            "manufacturer": "DENON PROFESSIONAL",
        },
        {
            "deviceType": "urn:schemas-denon-com:device:AiosDevice:1",
            "manufacturer": "Marantz",
        },
    ],
    "directv": [
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "manufacturer": "DIRECTV",
        },
    ],
    "dlna_dmr": [
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:1",
            "st": "urn:schemas-upnp-org:device:MediaRenderer:1",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:2",
            "st": "urn:schemas-upnp-org:device:MediaRenderer:2",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaRenderer:3",
            "st": "urn:schemas-upnp-org:device:MediaRenderer:3",
        },
    ],
    "dlna_dms": [
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:1",
            "st": "urn:schemas-upnp-org:device:MediaServer:1",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:2",
            "st": "urn:schemas-upnp-org:device:MediaServer:2",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:3",
            "st": "urn:schemas-upnp-org:device:MediaServer:3",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:MediaServer:4",
            "st": "urn:schemas-upnp-org:device:MediaServer:4",
        },
    ],
    "fritz": [
        {
            "st": "urn:schemas-upnp-org:device:fritzbox:1",
        },
    ],
    "fritzbox": [
        {
            "st": "urn:schemas-upnp-org:device:fritzbox:1",
        },
    ],
    "frontier_silicon": [
        {
            "st": "urn:schemas-frontier-silicon-com:undok:fsapi:1",
        },
    ],
    "harmony": [
        {
            "deviceType": "urn:myharmony-com:device:harmony:1",
            "manufacturer": "Logitech",
        },
    ],
    "heos": [
        {
            "st": "urn:schemas-denon-com:device:ACT-Denon:1",
        },
    ],
    "huawei_lte": [
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "Huawei",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "Huawei Technologies Co., Ltd.",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "SOYEA TECHNOLOGY CO., LTD.",
        },
    ],
    "hyperion": [
        {
            "manufacturer": "Hyperion Open Source Ambient Lighting",
            "st": "urn:hyperion-project.org:device:basic:1",
        },
    ],
    "isy994": [
        {
            "deviceType": "urn:udi-com:device:X_Insteon_Lighting_Device:1",
            "manufacturer": "Universal Devices Inc.",
        },
    ],
    "kaleidescape": [
        {
            "deviceType": "schemas-upnp-org:device:Basic:1",
            "manufacturer": "Kaleidescape, Inc.",
        },
    ],
    "keenetic_ndms2": [
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "Keenetic Ltd.",
        },
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "ZyXEL Communications Corp.",
        },
    ],
    "konnected": [
        {
            "manufacturer": "konnected.io",
        },
    ],
    "lametric": [
        {
            "deviceType": "urn:schemas-upnp-org:device:LaMetric:1",
        },
    ],
    "nanoleaf": [
        {
            "st": "Nanoleaf_aurora:light",
        },
        {
            "st": "nanoleaf:nl29",
        },
        {
            "st": "nanoleaf:nl42",
        },
        {
            "st": "nanoleaf:nl52",
        },
    ],
    "netgear": [
        {
            "deviceType": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
            "manufacturer": "NETGEAR, Inc.",
        },
    ],
    "octoprint": [
        {
            "deviceType": "urn:schemas-upnp-org:device:Basic:1",
            "manufacturer": "The OctoPrint Project",
        },
    ],
    "openhome": [
        {
            "st": "urn:av-openhome-org:service:Product:1",
        },
        {
            "st": "urn:av-openhome-org:service:Product:2",
        },
        {
            "st": "urn:av-openhome-org:service:Product:3",
        },
        {
            "st": "urn:av-openhome-org:service:Product:4",
        },
    ],
    "roku": [
        {
            "deviceType": "urn:roku-com:device:player:1-0",
            "manufacturer": "Roku",
            "st": "roku:ecp",
        },
    ],
    "samsungtv": [
        {
            "st": "urn:samsung.com:device:RemoteControlReceiver:1",
        },
        {
            "st": "urn:samsung.com:service:MainTVAgent2:1",
        },
        {
            "manufacturer": "Samsung",
            "st": "urn:schemas-upnp-org:service:RenderingControl:1",
        },
        {
            "manufacturer": "Samsung Electronics",
            "st": "urn:schemas-upnp-org:service:RenderingControl:1",
        },
    ],
    "songpal": [
        {
            "manufacturer": "Sony Corporation",
            "st": "urn:schemas-sony-com:service:ScalarWebAPI:1",
        },
    ],
    "sonos": [
        {
            "st": "urn:schemas-upnp-org:device:ZonePlayer:1",
        },
    ],
    "syncthru": [
        {
            "deviceType": "urn:schemas-upnp-org:device:Printer:1",
            "manufacturer": "Samsung Electronics",
        },
    ],
    "synology_dsm": [
        {
            "deviceType": "urn:schemas-upnp-org:device:Basic:1",
            "manufacturer": "Synology",
        },
    ],
    "unifi": [
        {
            "manufacturer": "Ubiquiti Networks",
            "modelDescription": "UniFi Dream Machine",
        },
        {
            "manufacturer": "Ubiquiti Networks",
            "modelDescription": "UniFi Dream Machine Pro",
        },
        {
            "manufacturer": "Ubiquiti Networks",
            "modelDescription": "UniFi Dream Machine SE",
        },
    ],
    "unifiprotect": [
        {
            "manufacturer": "Ubiquiti Networks",
            "modelDescription": "UniFi Dream Machine",
        },
        {
            "manufacturer": "Ubiquiti Networks",
            "modelDescription": "UniFi Dream Machine Pro",
        },
        {
            "manufacturer": "Ubiquiti Networks",
            "modelDescription": "UniFi Dream Machine SE",
        },
    ],
    "upnp": [
        {
            "st": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
        },
        {
            "st": "urn:schemas-upnp-org:device:InternetGatewayDevice:2",
        },
        {
            "nt": "urn:schemas-upnp-org:device:InternetGatewayDevice:1",
        },
        {
            "nt": "urn:schemas-upnp-org:device:InternetGatewayDevice:2",
        },
    ],
    "webostv": [
        {
            "st": "urn:lge-com:service:webos-second-screen:1",
        },
    ],
    "wemo": [
        {
            "manufacturer": "Belkin International Inc.",
        },
    ],
    "wilight": [
        {
            "manufacturer": "All Automacao Ltda",
        },
    ],
    "yamaha_musiccast": [
        {
            "manufacturer": "Yamaha Corporation",
        },
    ],
}
