"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

from __future__ import annotations

BLUETOOTH: list[dict[str, bool | str | int | list[int]]] = [
    {
        "domain": "airthings_ble",
        "manufacturer_id": 820,
    },
    {
        "connectable": False,
        "domain": "aranet",
        "manufacturer_id": 1794,
        "service_uuid": "f0cd1400-95da-4f4b-9ac8-aa55d312af0c",
    },
    {
        "connectable": False,
        "domain": "aranet",
        "manufacturer_id": 1794,
        "service_uuid": "0000fce0-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "bluemaestro",
        "manufacturer_id": 307,
    },
    {
        "connectable": False,
        "domain": "bthome",
        "service_data_uuid": "0000181c-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "bthome",
        "service_data_uuid": "0000181e-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "bthome",
        "service_data_uuid": "0000fcd2-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "dormakaba_dkey",
        "service_uuid": "e7a60000-6639-429f-94fd-86de8ea26897",
    },
    {
        "domain": "dormakaba_dkey",
        "service_uuid": "e7a60001-6639-429f-94fd-86de8ea26897",
    },
    {
        "domain": "eufylife_ble",
        "local_name": "eufy T9140",
    },
    {
        "domain": "eufylife_ble",
        "local_name": "eufy T9146",
    },
    {
        "domain": "eufylife_ble",
        "local_name": "eufy T9147",
    },
    {
        "domain": "eufylife_ble",
        "local_name": "eufy T9148",
    },
    {
        "domain": "eufylife_ble",
        "local_name": "eufy T9149",
    },
    {
        "connectable": False,
        "domain": "fjaraskupan",
        "manufacturer_data_start": [
            79,
            68,
            70,
            74,
            65,
            82,
        ],
        "manufacturer_id": 20296,
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "local_name": "Govee*",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "local_name": "GVH5*",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "local_name": "B5178*",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 6966,
        "service_uuid": "00008451-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 63391,
        "service_uuid": "00008351-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 26589,
        "service_uuid": "00008351-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 57391,
        "service_uuid": "00008351-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 18994,
        "service_uuid": "00008551-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 818,
        "service_uuid": "00008551-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 53579,
        "service_uuid": "00008151-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 43682,
        "service_uuid": "00008151-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 59970,
        "service_uuid": "00008151-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 63585,
        "service_uuid": "00008151-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 14474,
        "service_uuid": "00008151-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 10032,
        "service_uuid": "00008251-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "govee_ble",
        "manufacturer_id": 19506,
        "service_uuid": "00001801-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "homekit_controller",
        "manufacturer_data_start": [
            6,
        ],
        "manufacturer_id": 76,
    },
    {
        "domain": "ibeacon",
        "manufacturer_data_start": [
            2,
            21,
        ],
        "manufacturer_id": 76,
    },
    {
        "connectable": False,
        "domain": "inkbird",
        "local_name": "sps",
    },
    {
        "connectable": False,
        "domain": "inkbird",
        "local_name": "Inkbird*",
    },
    {
        "connectable": False,
        "domain": "inkbird",
        "local_name": "iBBQ*",
    },
    {
        "connectable": False,
        "domain": "inkbird",
        "local_name": "xBBQ*",
    },
    {
        "connectable": False,
        "domain": "inkbird",
        "local_name": "tps",
    },
    {
        "connectable": False,
        "domain": "kegtron",
        "manufacturer_id": 65535,
    },
    {
        "domain": "keymitt_ble",
        "service_uuid": "0000abcd-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "keymitt_ble",
        "local_name": "mib*",
    },
    {
        "domain": "ld2410_ble",
        "local_name": "HLK-LD2410B_*",
    },
    {
        "domain": "ld2410_ble",
        "local_name": "HLK-LD2410_*",
    },
    {
        "domain": "ld2410_ble",
        "manufacturer_data_start": [
            7,
            1,
        ],
        "manufacturer_id": 256,
        "service_uuid": "0000af30-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "led_ble",
        "local_name": "LEDnet*",
    },
    {
        "domain": "led_ble",
        "local_name": "BLE-LED*",
    },
    {
        "domain": "led_ble",
        "local_name": "LEDBLE*",
    },
    {
        "domain": "led_ble",
        "local_name": "Triones*",
    },
    {
        "domain": "led_ble",
        "local_name": "LEDBlue*",
    },
    {
        "domain": "led_ble",
        "local_name": "Dream~*",
    },
    {
        "domain": "led_ble",
        "local_name": "QHM-*",
    },
    {
        "domain": "led_ble",
        "local_name": "AP-*",
    },
    {
        "domain": "melnor",
        "manufacturer_data_start": [
            89,
        ],
        "manufacturer_id": 13,
    },
    {
        "connectable": False,
        "domain": "moat",
        "local_name": "Moat_S*",
    },
    {
        "connectable": False,
        "domain": "mopeka",
        "manufacturer_data_start": [
            3,
        ],
        "manufacturer_id": 89,
        "service_uuid": "0000fee5-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "mopeka",
        "manufacturer_data_start": [
            8,
        ],
        "manufacturer_id": 89,
        "service_uuid": "0000fee5-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "oralb",
        "manufacturer_id": 220,
    },
    {
        "connectable": False,
        "domain": "qingping",
        "local_name": "Qingping*",
    },
    {
        "connectable": False,
        "domain": "qingping",
        "local_name": "Lee Guitars*",
    },
    {
        "connectable": False,
        "domain": "qingping",
        "service_data_uuid": "0000fdcd-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "rapt_ble",
        "manufacturer_data_start": [
            80,
            84,
        ],
        "manufacturer_id": 16722,
    },
    {
        "domain": "rapt_ble",
        "manufacturer_data_start": [
            71,
        ],
        "manufacturer_id": 17739,
    },
    {
        "connectable": False,
        "domain": "ruuvitag_ble",
        "manufacturer_id": 1177,
    },
    {
        "connectable": False,
        "domain": "ruuvitag_ble",
        "local_name": "Ruuvi *",
    },
    {
        "connectable": False,
        "domain": "sensirion_ble",
        "manufacturer_id": 1749,
    },
    {
        "connectable": False,
        "domain": "sensirion_ble",
        "local_name": "MyCO2*",
    },
    {
        "connectable": False,
        "domain": "sensorpro",
        "manufacturer_data_start": [
            1,
            1,
            164,
            193,
        ],
        "manufacturer_id": 43605,
    },
    {
        "connectable": False,
        "domain": "sensorpro",
        "manufacturer_data_start": [
            1,
            5,
            164,
            193,
        ],
        "manufacturer_id": 43605,
    },
    {
        "connectable": False,
        "domain": "sensorpush",
        "local_name": "SensorPush*",
    },
    {
        "domain": "snooz",
        "local_name": "Snooz*",
    },
    {
        "domain": "snooz",
        "service_uuid": "729f0608-496a-47fe-a124-3a62aaa3fbc0",
    },
    {
        "connectable": False,
        "domain": "switchbot",
        "service_data_uuid": "00000d00-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "switchbot",
        "service_data_uuid": "0000fd3d-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "switchbot",
        "service_uuid": "cba20d00-224d-11e6-9fb8-0002a5d5c51b",
    },
    {
        "connectable": False,
        "domain": "switchbot",
        "manufacturer_id": 2409,
    },
    {
        "connectable": True,
        "domain": "switchbot",
        "manufacturer_id": 89,
    },
    {
        "connectable": True,
        "domain": "switchbot",
        "manufacturer_id": 741,
    },
    {
        "connectable": False,
        "domain": "thermobeacon",
        "manufacturer_data_start": [
            0,
        ],
        "manufacturer_id": 16,
        "service_uuid": "0000fff0-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "thermobeacon",
        "manufacturer_data_start": [
            0,
        ],
        "manufacturer_id": 17,
        "service_uuid": "0000fff0-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "thermobeacon",
        "manufacturer_data_start": [
            0,
        ],
        "manufacturer_id": 21,
        "service_uuid": "0000fff0-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "thermobeacon",
        "manufacturer_data_start": [
            0,
        ],
        "manufacturer_id": 24,
        "service_uuid": "0000fff0-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "thermobeacon",
        "manufacturer_data_start": [
            0,
        ],
        "manufacturer_id": 27,
        "service_uuid": "0000fff0-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "thermobeacon",
        "local_name": "ThermoBeacon",
    },
    {
        "connectable": False,
        "domain": "thermopro",
        "local_name": "TP35*",
    },
    {
        "connectable": False,
        "domain": "thermopro",
        "local_name": "TP39*",
    },
    {
        "domain": "tilt_ble",
        "manufacturer_data_start": [
            2,
            21,
            164,
            149,
            187,
        ],
        "manufacturer_id": 76,
    },
    {
        "connectable": False,
        "domain": "xiaomi_ble",
        "service_data_uuid": "0000fd50-0000-1000-8000-00805f9b34fb",
    },
    {
        "connectable": False,
        "domain": "xiaomi_ble",
        "service_data_uuid": "0000fe95-0000-1000-8000-00805f9b34fb",
    },
    {
        "domain": "yalexs_ble",
        "manufacturer_id": 465,
        "service_uuid": "0000fe24-0000-1000-8000-00805f9b34fb",
    },
]
