"""Automatically generated file.

To update, run python3 -m script.microsoft_tts
"""

SUPPORTED_LANGUAGES = {
    "af-za",
    "am-et",
    "ar-ae",
    "ar-bh",
    "ar-dz",
    "ar-eg",
    "ar-iq",
    "ar-jo",
    "ar-kw",
    "ar-lb",
    "ar-ly",
    "ar-ma",
    "ar-om",
    "ar-qa",
    "ar-sa",
    "ar-sy",
    "ar-tn",
    "ar-ye",
    "az-az",
    "bg-bg",
    "bn-bd",
    "bn-in",
    "bs-ba",
    "ca-es",
    "cs-cz",
    "cy-gb",
    "da-dk",
    "de-at",
    "de-ch",
    "de-de",
    "el-gr",
    "en-au",
    "en-ca",
    "en-gb",
    "en-hk",
    "en-ie",
    "en-in",
    "en-ke",
    "en-ng",
    "en-nz",
    "en-ph",
    "en-sg",
    "en-tz",
    "en-us",
    "en-za",
    "es-ar",
    "es-bo",
    "es-cl",
    "es-co",
    "es-cr",
    "es-cu",
    "es-do",
    "es-ec",
    "es-es",
    "es-gq",
    "es-gt",
    "es-hn",
    "es-mx",
    "es-ni",
    "es-pa",
    "es-pe",
    "es-pr",
    "es-py",
    "es-sv",
    "es-us",
    "es-uy",
    "es-ve",
    "et-ee",
    "eu-es",
    "fa-ir",
    "fi-fi",
    "fil-ph",
    "fr-be",
    "fr-ca",
    "fr-ch",
    "fr-fr",
    "ga-ie",
    "gl-es",
    "gu-in",
    "he-il",
    "hi-in",
    "hr-hr",
    "hu-hu",
    "hy-am",
    "id-id",
    "is-is",
    "it-it",
    "ja-jp",
    "jv-id",
    "ka-ge",
    "kk-kz",
    "km-kh",
    "kn-in",
    "ko-kr",
    "lo-la",
    "lt-lt",
    "lv-lv",
    "mk-mk",
    "ml-in",
    "mn-mn",
    "mr-in",
    "ms-my",
    "mt-mt",
    "my-mm",
    "nb-no",
    "ne-np",
    "nl-be",
    "nl-nl",
    "pl-pl",
    "ps-af",
    "pt-br",
    "pt-pt",
    "ro-ro",
    "ru-ru",
    "si-lk",
    "sk-sk",
    "sl-si",
    "so-so",
    "sq-al",
    "sr-rs",
    "su-id",
    "sv-se",
    "sw-ke",
    "sw-tz",
    "ta-in",
    "ta-lk",
    "ta-my",
    "ta-sg",
    "te-in",
    "th-th",
    "tr-tr",
    "uk-ua",
    "ur-in",
    "ur-pk",
    "uz-uz",
    "vi-vn",
    "wuu-cn",
    "yue-cn",
    "zh-cn",
    "zh-cn-henan",
    "zh-cn-liaoning",
    "zh-cn-shaanxi",
    "zh-cn-shandong",
    "zh-cn-sichuan",
    "zh-hk",
    "zh-tw",
    "zu-za",
}
