"""Constants for the Coolmaster integration."""

DATA_INFO = "info"
DATA_COORDINATOR = "coordinator"

DOMAIN = "coolmaster"

DEFAULT_PORT = 10102

CONF_SUPPORTED_MODES = "supported_modes"
CONF_SWING_SUPPORT = "swing_support"
