"""Constants for the Electra Air Conditioner integration."""

DOMAIN = "electrasmart"

CONF_PHONE_NUMBER = "phone_number"
CONF_OTP = "one_time_password"
CONF_IMEI = "imei"
SCAN_INTERVAL_SEC = 30
API_DELAY = 5
CONSECUTIVE_FAILURE_THRESHOLD = 4
UNAVAILABLE_THRESH_SEC = 120
PRESET_NONE = "None"
PRESET_SHABAT = "Shabat"
