"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

APPLICATION_CREDENTIALS = [
    "geocaching",
    "google",
    "google_assistant_sdk",
    "google_mail",
    "google_sheets",
    "home_connect",
    "lametric",
    "lyric",
    "neato",
    "nest",
    "netatmo",
    "senz",
    "spotify",
    "withings",
    "xbox",
    "yolink",
    "youtube",
]
