"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

from __future__ import annotations

DHCP: list[dict[str, str | bool]] = [
    {
        "domain": "airzone",
        "macaddress": "E84F25*",
    },
    {
        "domain": "august",
        "hostname": "yale-connect-plus",
        "macaddress": "00177A*",
    },
    {
        "domain": "august",
        "hostname": "connect",
        "macaddress": "D86162*",
    },
    {
        "domain": "august",
        "hostname": "connect",
        "macaddress": "B8B7F1*",
    },
    {
        "domain": "august",
        "hostname": "connect",
        "macaddress": "2C9FFB*",
    },
    {
        "domain": "august",
        "hostname": "august*",
        "macaddress": "E076D0*",
    },
    {
        "domain": "awair",
        "macaddress": "70886B1*",
    },
    {
        "domain": "axis",
        "registered_devices": True,
    },
    {
        "domain": "axis",
        "hostname": "axis-00408c*",
        "macaddress": "00408C*",
    },
    {
        "domain": "axis",
        "hostname": "axis-accc8e*",
        "macaddress": "ACCC8E*",
    },
    {
        "domain": "axis",
        "hostname": "axis-b8a44f*",
        "macaddress": "B8A44F*",
    },
    {
        "domain": "blink",
        "hostname": "blink*",
        "macaddress": "B85F98*",
    },
    {
        "domain": "blink",
        "hostname": "blink*",
        "macaddress": "00037F*",
    },
    {
        "domain": "blink",
        "hostname": "blink*",
        "macaddress": "20A171*",
    },
    {
        "domain": "broadlink",
        "registered_devices": True,
    },
    {
        "domain": "broadlink",
        "macaddress": "34EA34*",
    },
    {
        "domain": "broadlink",
        "macaddress": "24DFA7*",
    },
    {
        "domain": "broadlink",
        "macaddress": "A043B0*",
    },
    {
        "domain": "broadlink",
        "macaddress": "B4430D*",
    },
    {
        "domain": "broadlink",
        "macaddress": "C8F742*",
    },
    {
        "domain": "broadlink",
        "macaddress": "E81656*",
    },
    {
        "domain": "broadlink",
        "macaddress": "E87072*",
    },
    {
        "domain": "broadlink",
        "macaddress": "EC0BAE*",
    },
    {
        "domain": "dlink",
        "hostname": "dsp-w215",
    },
    {
        "domain": "elkm1",
        "registered_devices": True,
    },
    {
        "domain": "elkm1",
        "macaddress": "00409D*",
    },
    {
        "domain": "emonitor",
        "hostname": "emonitor*",
        "macaddress": "0090C2*",
    },
    {
        "domain": "emonitor",
        "registered_devices": True,
    },
    {
        "domain": "esphome",
        "registered_devices": True,
    },
    {
        "domain": "flume",
        "hostname": "flume-gw-*",
    },
    {
        "domain": "flux_led",
        "registered_devices": True,
    },
    {
        "domain": "flux_led",
        "hostname": "[ba][lk]*",
        "macaddress": "18B905*",
    },
    {
        "domain": "flux_led",
        "hostname": "[ba][lk]*",
        "macaddress": "249494*",
    },
    {
        "domain": "flux_led",
        "hostname": "[ba][lk]*",
        "macaddress": "7CB94C*",
    },
    {
        "domain": "flux_led",
        "hostname": "[hba][flk]*",
        "macaddress": "ACCF23*",
    },
    {
        "domain": "flux_led",
        "hostname": "[ba][lk]*",
        "macaddress": "B4E842*",
    },
    {
        "domain": "flux_led",
        "hostname": "[hba][flk]*",
        "macaddress": "F0FE6B*",
    },
    {
        "domain": "flux_led",
        "hostname": "lwip*",
        "macaddress": "8CCE4E*",
    },
    {
        "domain": "flux_led",
        "hostname": "hf-lpb100-zj*",
    },
    {
        "domain": "flux_led",
        "hostname": "zengge_[0-9a-f][0-9a-f]_*",
    },
    {
        "domain": "flux_led",
        "hostname": "sta*",
        "macaddress": "C82E47*",
    },
    {
        "domain": "fronius",
        "macaddress": "0003AC*",
    },
    {
        "domain": "fully_kiosk",
        "registered_devices": True,
    },
    {
        "domain": "goalzero",
        "registered_devices": True,
    },
    {
        "domain": "goalzero",
        "hostname": "yeti*",
    },
    {
        "domain": "gogogate2",
        "hostname": "ismartgate*",
    },
    {
        "domain": "guardian",
        "hostname": "gvc*",
        "macaddress": "30AEA4*",
    },
    {
        "domain": "guardian",
        "hostname": "gvc*",
        "macaddress": "B4E62D*",
    },
    {
        "domain": "guardian",
        "hostname": "guardian*",
        "macaddress": "30AEA4*",
    },
    {
        "domain": "hunterdouglas_powerview",
        "registered_devices": True,
    },
    {
        "domain": "hunterdouglas_powerview",
        "hostname": "hunter*",
        "macaddress": "002674*",
    },
    {
        "domain": "insteon",
        "macaddress": "000EF3*",
    },
    {
        "domain": "insteon",
        "registered_devices": True,
    },
    {
        "domain": "intellifire",
        "hostname": "zentrios-*",
    },
    {
        "domain": "isy994",
        "registered_devices": True,
    },
    {
        "domain": "isy994",
        "hostname": "isy*",
        "macaddress": "0021B9*",
    },
    {
        "domain": "isy994",
        "hostname": "eisy*",
        "macaddress": "0021B9*",
    },
    {
        "domain": "isy994",
        "hostname": "polisy*",
        "macaddress": "000DB9*",
    },
    {
        "domain": "lametric",
        "registered_devices": True,
    },
    {
        "domain": "lifx",
        "macaddress": "D073D5*",
    },
    {
        "domain": "lifx",
        "registered_devices": True,
    },
    {
        "domain": "litterrobot",
        "hostname": "litter-robot4",
    },
    {
        "domain": "lyric",
        "hostname": "lyric-*",
        "macaddress": "48A2E6*",
    },
    {
        "domain": "lyric",
        "hostname": "lyric-*",
        "macaddress": "B82CA0*",
    },
    {
        "domain": "lyric",
        "hostname": "lyric-*",
        "macaddress": "00D02D*",
    },
    {
        "domain": "motion_blinds",
        "registered_devices": True,
    },
    {
        "domain": "motion_blinds",
        "hostname": "motion_*",
    },
    {
        "domain": "motion_blinds",
        "hostname": "brel_*",
    },
    {
        "domain": "motion_blinds",
        "hostname": "connector_*",
    },
    {
        "domain": "myq",
        "macaddress": "645299*",
    },
    {
        "domain": "nest",
        "macaddress": "18B430*",
    },
    {
        "domain": "nest",
        "macaddress": "641666*",
    },
    {
        "domain": "nest",
        "macaddress": "D8EB46*",
    },
    {
        "domain": "nexia",
        "hostname": "xl857-*",
        "macaddress": "000231*",
    },
    {
        "domain": "nuheat",
        "hostname": "nuheat",
        "macaddress": "002338*",
    },
    {
        "domain": "nuki",
        "hostname": "nuki_bridge_*",
    },
    {
        "domain": "obihai",
        "macaddress": "9CADEF*",
    },
    {
        "domain": "oncue",
        "hostname": "kohlergen*",
        "macaddress": "00146F*",
    },
    {
        "domain": "onvif",
        "registered_devices": True,
    },
    {
        "domain": "overkiz",
        "hostname": "gateway*",
        "macaddress": "F8811A*",
    },
    {
        "domain": "powerwall",
        "hostname": "1118431-*",
    },
    {
        "domain": "powerwall",
        "hostname": "1232100-*",
    },
    {
        "domain": "prusalink",
        "macaddress": "109C70*",
    },
    {
        "domain": "qnap_qsw",
        "macaddress": "245EBE*",
    },
    {
        "domain": "rachio",
        "hostname": "rachio-*",
        "macaddress": "009D6B*",
    },
    {
        "domain": "rachio",
        "hostname": "rachio-*",
        "macaddress": "F0038C*",
    },
    {
        "domain": "rachio",
        "hostname": "rachio-*",
        "macaddress": "74C63B*",
    },
    {
        "domain": "radiotherm",
        "hostname": "thermostat*",
        "macaddress": "5CDAD4*",
    },
    {
        "domain": "radiotherm",
        "registered_devices": True,
    },
    {
        "domain": "rainforest_eagle",
        "macaddress": "D8D5B9*",
    },
    {
        "domain": "reolink",
        "hostname": "reolink*",
    },
    {
        "domain": "reolink",
        "macaddress": "EC71DB*",
    },
    {
        "domain": "reolink",
        "registered_devices": True,
    },
    {
        "domain": "ring",
        "hostname": "ring*",
        "macaddress": "0CAE7D*",
    },
    {
        "domain": "roomba",
        "hostname": "irobot-*",
        "macaddress": "501479*",
    },
    {
        "domain": "roomba",
        "hostname": "roomba-*",
        "macaddress": "80A589*",
    },
    {
        "domain": "roomba",
        "hostname": "roomba-*",
        "macaddress": "DCF505*",
    },
    {
        "domain": "roomba",
        "hostname": "roomba-*",
        "macaddress": "204EF6*",
    },
    {
        "domain": "ruuvi_gateway",
        "hostname": "ruuvigateway*",
    },
    {
        "domain": "samsungtv",
        "registered_devices": True,
    },
    {
        "domain": "samsungtv",
        "hostname": "tizen*",
    },
    {
        "domain": "samsungtv",
        "macaddress": "4844F7*",
    },
    {
        "domain": "samsungtv",
        "macaddress": "606BBD*",
    },
    {
        "domain": "samsungtv",
        "macaddress": "641CB0*",
    },
    {
        "domain": "samsungtv",
        "macaddress": "8CC8CD*",
    },
    {
        "domain": "samsungtv",
        "macaddress": "8CEA48*",
    },
    {
        "domain": "samsungtv",
        "macaddress": "F47B5E*",
    },
    {
        "domain": "screenlogic",
        "registered_devices": True,
    },
    {
        "domain": "screenlogic",
        "hostname": "pentair*",
        "macaddress": "00C033*",
    },
    {
        "domain": "sense",
        "hostname": "sense-*",
        "macaddress": "009D6B*",
    },
    {
        "domain": "sense",
        "hostname": "sense-*",
        "macaddress": "DCEFCA*",
    },
    {
        "domain": "sense",
        "hostname": "sense-*",
        "macaddress": "A4D578*",
    },
    {
        "domain": "sensibo",
        "hostname": "sensibo*",
    },
    {
        "domain": "simplisafe",
        "hostname": "simplisafe*",
        "macaddress": "30AEA4*",
    },
    {
        "domain": "sleepiq",
        "macaddress": "64DBA0*",
    },
    {
        "domain": "smartthings",
        "hostname": "st*",
        "macaddress": "24FD5B*",
    },
    {
        "domain": "smartthings",
        "hostname": "smartthings*",
        "macaddress": "24FD5B*",
    },
    {
        "domain": "smartthings",
        "hostname": "hub*",
        "macaddress": "24FD5B*",
    },
    {
        "domain": "smartthings",
        "hostname": "hub*",
        "macaddress": "D052A8*",
    },
    {
        "domain": "smartthings",
        "hostname": "hub*",
        "macaddress": "286D97*",
    },
    {
        "domain": "solaredge",
        "hostname": "target",
        "macaddress": "002702*",
    },
    {
        "domain": "somfy_mylink",
        "hostname": "somfy_*",
        "macaddress": "B8B7F1*",
    },
    {
        "domain": "squeezebox",
        "hostname": "squeezebox*",
        "macaddress": "000420*",
    },
    {
        "domain": "steamist",
        "registered_devices": True,
    },
    {
        "domain": "steamist",
        "hostname": "my[45]50*",
        "macaddress": "001E0C*",
    },
    {
        "domain": "tado",
        "hostname": "tado*",
    },
    {
        "domain": "tesla_wall_connector",
        "hostname": "teslawallconnector_*",
        "macaddress": "DC44271*",
    },
    {
        "domain": "tesla_wall_connector",
        "hostname": "teslawallconnector_*",
        "macaddress": "98ED5C*",
    },
    {
        "domain": "tesla_wall_connector",
        "hostname": "teslawallconnector_*",
        "macaddress": "4CFCAA*",
    },
    {
        "domain": "tolo",
        "hostname": "usr-tcp232-ed2",
    },
    {
        "domain": "toon",
        "hostname": "eneco-*",
        "macaddress": "74C63B*",
    },
    {
        "domain": "tplink",
        "registered_devices": True,
    },
    {
        "domain": "tplink",
        "hostname": "es*",
        "macaddress": "54AF97*",
    },
    {
        "domain": "tplink",
        "hostname": "ep*",
        "macaddress": "E848B8*",
    },
    {
        "domain": "tplink",
        "hostname": "ep*",
        "macaddress": "1C61B4*",
    },
    {
        "domain": "tplink",
        "hostname": "ep*",
        "macaddress": "003192*",
    },
    {
        "domain": "tplink",
        "hostname": "hs*",
        "macaddress": "1C3BF3*",
    },
    {
        "domain": "tplink",
        "hostname": "hs*",
        "macaddress": "50C7BF*",
    },
    {
        "domain": "tplink",
        "hostname": "hs*",
        "macaddress": "68FF7B*",
    },
    {
        "domain": "tplink",
        "hostname": "hs*",
        "macaddress": "98DAC4*",
    },
    {
        "domain": "tplink",
        "hostname": "hs*",
        "macaddress": "B09575*",
    },
    {
        "domain": "tplink",
        "hostname": "hs*",
        "macaddress": "C006C3*",
    },
    {
        "domain": "tplink",
        "hostname": "lb*",
        "macaddress": "1C3BF3*",
    },
    {
        "domain": "tplink",
        "hostname": "lb*",
        "macaddress": "50C7BF*",
    },
    {
        "domain": "tplink",
        "hostname": "lb*",
        "macaddress": "68FF7B*",
    },
    {
        "domain": "tplink",
        "hostname": "lb*",
        "macaddress": "98DAC4*",
    },
    {
        "domain": "tplink",
        "hostname": "lb*",
        "macaddress": "B09575*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "60A4B7*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "005F67*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "1027F5*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "B0A7B9*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "403F8C*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "C0C9E3*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "909A4A*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "E848B8*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "003192*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "1C3BF3*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "50C7BF*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "68FF7B*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "98DAC4*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "B09575*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "C006C3*",
    },
    {
        "domain": "tplink",
        "hostname": "k[lp]*",
        "macaddress": "6C5AB0*",
    },
    {
        "domain": "tuya",
        "macaddress": "105A17*",
    },
    {
        "domain": "tuya",
        "macaddress": "10D561*",
    },
    {
        "domain": "tuya",
        "macaddress": "1869D8*",
    },
    {
        "domain": "tuya",
        "macaddress": "381F8D*",
    },
    {
        "domain": "tuya",
        "macaddress": "508A06*",
    },
    {
        "domain": "tuya",
        "macaddress": "68572D*",
    },
    {
        "domain": "tuya",
        "macaddress": "708976*",
    },
    {
        "domain": "tuya",
        "macaddress": "7CF666*",
    },
    {
        "domain": "tuya",
        "macaddress": "84E342*",
    },
    {
        "domain": "tuya",
        "macaddress": "D4A651*",
    },
    {
        "domain": "tuya",
        "macaddress": "D81F12*",
    },
    {
        "domain": "twinkly",
        "hostname": "twinkly_*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "B4FBE4*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "802AA8*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "F09FC2*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "68D79A*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "18E829*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "245A4C*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "784558*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "E063DA*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "265A4C*",
    },
    {
        "domain": "unifiprotect",
        "macaddress": "74ACB9*",
    },
    {
        "domain": "verisure",
        "macaddress": "0023C1*",
    },
    {
        "domain": "vicare",
        "macaddress": "B87424*",
    },
    {
        "domain": "wiz",
        "registered_devices": True,
    },
    {
        "domain": "wiz",
        "macaddress": "A8BB50*",
    },
    {
        "domain": "wiz",
        "macaddress": "D8A011*",
    },
    {
        "domain": "wiz",
        "macaddress": "444F8E*",
    },
    {
        "domain": "wiz",
        "macaddress": "6C2990*",
    },
    {
        "domain": "wiz",
        "hostname": "wiz_*",
    },
    {
        "domain": "yeelight",
        "hostname": "yeelink-*",
    },
]
