"""Constants for the D-Link Power Plug integration."""

ATTRIBUTION = "Data provided by D-Link"
ATTR_TOTAL_CONSUMPTION = "total_consumption"

CONF_USE_LEGACY_PROTOCOL = "use_legacy_protocol"

DEFAULT_NAME = "D-Link Smart Plug W215"
DEFAULT_USERNAME = "admin"
DOMAIN = "dlink"

MANUFACTURER = "D-Link"
