"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

HOMEKIT = {
    "3810X": {
        "always_discover": True,
        "domain": "roku",
    },
    "3820X": {
        "always_discover": True,
        "domain": "roku",
    },
    "4660X": {
        "always_discover": True,
        "domain": "roku",
    },
    "7820X": {
        "always_discover": True,
        "domain": "roku",
    },
    "819LMB": {
        "always_discover": True,
        "domain": "myq",
    },
    "AC02": {
        "always_discover": True,
        "domain": "tado",
    },
    "Abode": {
        "always_discover": True,
        "domain": "abode",
    },
    "BSB002": {
        "always_discover": False,
        "domain": "hue",
    },
    "C105X": {
        "always_discover": True,
        "domain": "roku",
    },
    "C135X": {
        "always_discover": True,
        "domain": "roku",
    },
    "EB-*": {
        "always_discover": True,
        "domain": "ecobee",
    },
    "Escea": {
        "always_discover": False,
        "domain": "escea",
    },
    "HHKBridge*": {
        "always_discover": True,
        "domain": "hive",
    },
    "Healty Home Coach": {
        "always_discover": True,
        "domain": "netatmo",
    },
    "Iota": {
        "always_discover": True,
        "domain": "abode",
    },
    "LIFX A19": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX BR30": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Beam": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Candle": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Clean": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Color": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX DLCOL": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX DLWW": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Dlight": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Downlight": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Filament": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX GU10": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Lightstrip": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Mini": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Nightvision": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Pls": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Plus": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Tile": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX White": {
        "always_discover": True,
        "domain": "lifx",
    },
    "LIFX Z": {
        "always_discover": True,
        "domain": "lifx",
    },
    "MYQ": {
        "always_discover": True,
        "domain": "myq",
    },
    "NL29": {
        "always_discover": False,
        "domain": "nanoleaf",
    },
    "NL42": {
        "always_discover": False,
        "domain": "nanoleaf",
    },
    "NL47": {
        "always_discover": False,
        "domain": "nanoleaf",
    },
    "NL48": {
        "always_discover": False,
        "domain": "nanoleaf",
    },
    "NL52": {
        "always_discover": False,
        "domain": "nanoleaf",
    },
    "NL59": {
        "always_discover": False,
        "domain": "nanoleaf",
    },
    "Netatmo Relay": {
        "always_discover": True,
        "domain": "netatmo",
    },
    "PowerView": {
        "always_discover": True,
        "domain": "hunterdouglas_powerview",
    },
    "Presence": {
        "always_discover": True,
        "domain": "netatmo",
    },
    "Rachio": {
        "always_discover": True,
        "domain": "rachio",
    },
    "SPK5": {
        "always_discover": True,
        "domain": "rainmachine",
    },
    "Sensibo": {
        "always_discover": True,
        "domain": "sensibo",
    },
    "Smart Bridge": {
        "always_discover": False,
        "domain": "lutron_caseta",
    },
    "Socket": {
        "always_discover": False,
        "domain": "wemo",
    },
    "TRADFRI": {
        "always_discover": True,
        "domain": "tradfri",
    },
    "Touch HD": {
        "always_discover": True,
        "domain": "rainmachine",
    },
    "Welcome": {
        "always_discover": True,
        "domain": "netatmo",
    },
    "Wemo": {
        "always_discover": False,
        "domain": "wemo",
    },
    "YL*": {
        "always_discover": False,
        "domain": "yeelight",
    },
    "ecobee*": {
        "always_discover": True,
        "domain": "ecobee",
    },
    "iSmartGate": {
        "always_discover": True,
        "domain": "gogogate2",
    },
    "iZone": {
        "always_discover": True,
        "domain": "izone",
    },
    "tado": {
        "always_discover": True,
        "domain": "tado",
    },
}

ZEROCONF = {
    "_Volumio._tcp.local.": [
        {
            "domain": "volumio",
        },
    ],
    "_airplay._tcp.local.": [
        {
            "domain": "apple_tv",
            "properties": {
                "model": "appletv*",
            },
        },
        {
            "domain": "apple_tv",
            "properties": {
                "model": "audioaccessory*",
            },
        },
        {
            "domain": "apple_tv",
            "properties": {
                "am": "airport*",
            },
        },
        {
            "domain": "samsungtv",
            "properties": {
                "manufacturer": "samsung*",
            },
        },
    ],
    "_airport._tcp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_amzn-alexa._tcp.local.": [
        {
            "domain": "roomba",
            "name": "irobot-*",
        },
        {
            "domain": "roomba",
            "name": "roomba-*",
        },
    ],
    "_androidtvremote2._tcp.local.": [
        {
            "domain": "androidtv_remote",
        },
    ],
    "_api._tcp.local.": [
        {
            "domain": "baf",
            "properties": {
                "model": "haiku*",
            },
        },
        {
            "domain": "baf",
            "properties": {
                "model": "i6*",
            },
        },
    ],
    "_api._udp.local.": [
        {
            "domain": "guardian",
        },
    ],
    "_appletv-v2._tcp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_axis-video._tcp.local.": [
        {
            "domain": "axis",
            "properties": {
                "macaddress": "00408c*",
            },
        },
        {
            "domain": "axis",
            "properties": {
                "macaddress": "accc8e*",
            },
        },
        {
            "domain": "axis",
            "properties": {
                "macaddress": "b8a44f*",
            },
        },
        {
            "domain": "doorbird",
            "properties": {
                "macaddress": "1ccae3*",
            },
        },
    ],
    "_bbxsrv._tcp.local.": [
        {
            "domain": "blebox",
        },
    ],
    "_bond._tcp.local.": [
        {
            "domain": "bond",
        },
    ],
    "_companion-link._tcp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_daap._tcp.local.": [
        {
            "domain": "forked_daapd",
        },
    ],
    "_dkapi._tcp.local.": [
        {
            "domain": "daikin",
        },
    ],
    "_dvl-deviceapi._tcp.local.": [
        {
            "domain": "devolo_home_control",
        },
        {
            "domain": "devolo_home_network",
            "properties": {
                "MT": "*",
            },
        },
    ],
    "_easylink._tcp.local.": [
        {
            "domain": "modern_forms",
            "name": "wac*",
        },
    ],
    "_elg._tcp.local.": [
        {
            "domain": "elgato",
        },
    ],
    "_enphase-envoy._tcp.local.": [
        {
            "domain": "enphase_envoy",
        },
    ],
    "_esphomelib._tcp.local.": [
        {
            "domain": "esphome",
        },
        {
            "domain": "zha",
            "name": "tube*",
        },
    ],
    "_fbx-api._tcp.local.": [
        {
            "domain": "freebox",
        },
    ],
    "_googlecast._tcp.local.": [
        {
            "domain": "cast",
        },
    ],
    "_hap._tcp.local.": [
        {
            "domain": "homekit_controller",
        },
        {
            "domain": "zwave_me",
            "name": "*z.wave-me*",
        },
    ],
    "_hap._udp.local.": [
        {
            "domain": "homekit_controller",
        },
    ],
    "_homekit._tcp.local.": [
        {
            "domain": "homekit",
        },
    ],
    "_hscp._tcp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_http._tcp.local.": [
        {
            "domain": "awair",
            "name": "awair*",
        },
        {
            "domain": "bosch_shc",
            "name": "bosch shc*",
        },
        {
            "domain": "loqed",
            "name": "loqed*",
        },
        {
            "domain": "nam",
            "name": "nam-*",
        },
        {
            "domain": "nam",
            "properties": {
                "manufacturer": "nettigo",
            },
        },
        {
            "domain": "pure_energie",
            "name": "smartbridge*",
        },
        {
            "domain": "rachio",
            "name": "rachio*",
        },
        {
            "domain": "rainmachine",
            "name": "rainmachine*",
        },
        {
            "domain": "shelly",
            "name": "shelly*",
        },
        {
            "domain": "synology_dsm",
            "properties": {
                "vendor": "synology*",
            },
        },
    ],
    "_hue._tcp.local.": [
        {
            "domain": "hue",
        },
    ],
    "_hwenergy._tcp.local.": [
        {
            "domain": "homewizard",
        },
    ],
    "_ipp._tcp.local.": [
        {
            "domain": "ipp",
        },
    ],
    "_ipps._tcp.local.": [
        {
            "domain": "ipp",
        },
    ],
    "_kizbox._tcp.local.": [
        {
            "domain": "overkiz",
            "name": "gateway*",
        },
    ],
    "_lookin._tcp.local.": [
        {
            "domain": "lookin",
        },
    ],
    "_lutron._tcp.local.": [
        {
            "domain": "lutron_caseta",
            "properties": {
                "SYSTYPE": "radiora3*",
            },
        },
        {
            "domain": "lutron_caseta",
            "properties": {
                "SYSTYPE": "smartbridge*",
            },
        },
        {
            "domain": "lutron_caseta",
            "properties": {
                "SYSTYPE": "ra2select*",
            },
        },
    ],
    "_mediaremotetv._tcp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_meshcop._udp.local.": [
        {
            "domain": "thread",
        },
    ],
    "_miio._udp.local.": [
        {
            "domain": "xiaomi_aqara",
        },
        {
            "domain": "xiaomi_miio",
        },
        {
            "domain": "yeelight",
            "name": "yeelink-*",
        },
    ],
    "_nanoleafapi._tcp.local.": [
        {
            "domain": "nanoleaf",
        },
    ],
    "_nanoleafms._tcp.local.": [
        {
            "domain": "nanoleaf",
        },
    ],
    "_nut._tcp.local.": [
        {
            "domain": "nut",
        },
    ],
    "_octoprint._tcp.local.": [
        {
            "domain": "octoprint",
        },
    ],
    "_plexmediasvr._tcp.local.": [
        {
            "domain": "plex",
        },
    ],
    "_plugwise._tcp.local.": [
        {
            "domain": "plugwise",
        },
    ],
    "_powerview._tcp.local.": [
        {
            "domain": "hunterdouglas_powerview",
        },
    ],
    "_printer._tcp.local.": [
        {
            "domain": "brother",
            "name": "brother*",
        },
    ],
    "_raop._tcp.local.": [
        {
            "domain": "apple_tv",
            "properties": {
                "am": "appletv*",
            },
        },
        {
            "domain": "apple_tv",
            "properties": {
                "am": "audioaccessory*",
            },
        },
        {
            "domain": "apple_tv",
            "properties": {
                "am": "airport*",
            },
        },
    ],
    "_sideplay._tcp.local.": [
        {
            "domain": "ecobee",
            "properties": {
                "mdl": "eb-*",
            },
        },
        {
            "domain": "ecobee",
            "properties": {
                "mdl": "ecobee*",
            },
        },
    ],
    "_sleep-proxy._udp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_slzb-06._tcp.local.": [
        {
            "domain": "zha",
            "name": "slzb-06*",
        },
    ],
    "_sonos._tcp.local.": [
        {
            "domain": "sonos",
        },
    ],
    "_soundtouch._tcp.local.": [
        {
            "domain": "soundtouch",
        },
    ],
    "_spotify-connect._tcp.local.": [
        {
            "domain": "spotify",
        },
    ],
    "_ssh._tcp.local.": [
        {
            "domain": "smappee",
            "name": "smappee1*",
        },
        {
            "domain": "smappee",
            "name": "smappee2*",
        },
        {
            "domain": "smappee",
            "name": "smappee50*",
        },
    ],
    "_system-bridge._tcp.local.": [
        {
            "domain": "system_bridge",
        },
    ],
    "_touch-able._tcp.local.": [
        {
            "domain": "apple_tv",
        },
    ],
    "_viziocast._tcp.local.": [
        {
            "domain": "vizio",
        },
    ],
    "_wled._tcp.local.": [
        {
            "domain": "wled",
        },
    ],
    "_xbmc-jsonrpc-h._tcp.local.": [
        {
            "domain": "kodi",
        },
    ],
    "_zigate-zigbee-gateway._tcp.local.": [
        {
            "domain": "zha",
            "name": "*zigate*",
        },
    ],
    "_zigstar_gw._tcp.local.": [
        {
            "domain": "zha",
            "name": "*zigstar*",
        },
    ],
    "_zwave-js-server._tcp.local.": [
        {
            "domain": "zwave_js",
        },
    ],
}
