"""Constants for the EDL21 component."""
import logging

LOGGER = logging.getLogger(__package__)

DOMAIN = "edl21"

CONF_SERIAL_PORT = "serial_port"

SIGNAL_EDL21_TELEGRAM = "edl21_telegram"

DEFAULT_TITLE = "Smart Meter"

DEFAULT_DEVICE_NAME = "Smart Meter"
