"""Provides the constants needed for the component."""

DOMAIN = "text"

ATTR_MAX = "max"
ATTR_MIN = "min"
ATTR_MODE = "mode"
ATTR_PATTERN = "pattern"
ATTR_VALUE = "value"

SERVICE_SET_VALUE = "set_value"
