"""Provides the constants needed for the component."""

DOMAIN = "select"

ATTR_CYCLE = "cycle"
ATTR_OPTIONS = "options"
ATTR_OPTION = "option"

CONF_CYCLE = "cycle"
CONF_OPTION = "option"

SERVICE_SELECT_FIRST = "select_first"
SERVICE_SELECT_LAST = "select_last"
SERVICE_SELECT_NEXT = "select_next"
SERVICE_SELECT_OPTION = "select_option"
SERVICE_SELECT_PREVIOUS = "select_previous"
