"""Constants for the Yale Access Bluetooth integration."""

DOMAIN = "yalexs_ble"

CONF_LOCAL_NAME = "local_name"
CONF_KEY = "key"
CONF_SLOT = "slot"
CONF_ALWAYS_CONNECTED = "always_connected"

DEVICE_TIMEOUT = 55
