"""Constants for the Frontier Silicon Media Player integration."""
DOMAIN = "frontier_silicon"

CONF_WEBFSAPI_URL = "webfsapi_url"
CONF_PIN = "pin"

SSDP_ST = "urn:schemas-frontier-silicon-com:undok:fsapi:1"
SSDP_ATTR_SPEAKER_NAME = "SPEAKER-NAME"

DEFAULT_PIN = "1234"
DEFAULT_PORT = 80

MEDIA_CONTENT_ID_PRESET = "preset"
MEDIA_CONTENT_ID_CHANNELS = "channels"
