"""Automatically generated file.

To update, run python3 -m script.hassfest
"""

USB = [
    {
        "description": "*skyconnect v1.0*",
        "domain": "homeassistant_sky_connect",
        "pid": "EA60",
        "vid": "10C4",
    },
    {
        "domain": "insteon",
        "vid": "10BF",
    },
    {
        "domain": "modem_callerid",
        "pid": "1340",
        "vid": "0572",
    },
    {
        "domain": "velbus",
        "pid": "0B1B",
        "vid": "10CF",
    },
    {
        "domain": "velbus",
        "pid": "0516",
        "vid": "10CF",
    },
    {
        "domain": "velbus",
        "pid": "0517",
        "vid": "10CF",
    },
    {
        "domain": "velbus",
        "pid": "0518",
        "vid": "10CF",
    },
    {
        "description": "*2652*",
        "domain": "zha",
        "pid": "EA60",
        "vid": "10C4",
    },
    {
        "description": "*sonoff*plus*",
        "domain": "zha",
        "pid": "55D4",
        "vid": "1A86",
    },
    {
        "description": "*sonoff*plus*",
        "domain": "zha",
        "pid": "EA60",
        "vid": "10C4",
    },
    {
        "description": "*tubeszb*",
        "domain": "zha",
        "pid": "EA60",
        "vid": "10C4",
    },
    {
        "description": "*tubeszb*",
        "domain": "zha",
        "pid": "7523",
        "vid": "1A86",
    },
    {
        "description": "*zigstar*",
        "domain": "zha",
        "pid": "7523",
        "vid": "1A86",
    },
    {
        "description": "*conbee*",
        "domain": "zha",
        "pid": "0030",
        "vid": "1CF1",
    },
    {
        "description": "*zigbee*",
        "domain": "zha",
        "pid": "8A2A",
        "vid": "10C4",
    },
    {
        "description": "*zigate*",
        "domain": "zha",
        "pid": "6015",
        "vid": "0403",
    },
    {
        "description": "*zigate*",
        "domain": "zha",
        "pid": "EA60",
        "vid": "10C4",
    },
    {
        "description": "*bv 2010/10*",
        "domain": "zha",
        "pid": "8B34",
        "vid": "10C4",
    },
    {
        "domain": "zwave_js",
        "pid": "0200",
        "vid": "0658",
    },
    {
        "description": "*z-wave*",
        "domain": "zwave_js",
        "pid": "8A2A",
        "vid": "10C4",
    },
]
